import {
    Client,
    Query,
    Mutation,
    ArgsQuery,
    ArgsMutation,
    ClientConfig,
} from '@/types'
import { BaseAPI } from './BaseAPI';
import { ApplicationError, ValidationError} from './Error'

enum Operation {
    QUERY = 'query',
    MUTATION = 'mutation'
}

type OperationName = keyof Query | keyof Mutation

type ValueOf<T> = T[keyof T];

type ValuesOf<T>= T extends any[] ? T[number] : T;

type BodyChild<T> = {
    [Key in keyof T]?: Body<T[Key]>
}

type Body<T> = Array<keyof NonNullable<ValuesOf<T>> | BodyChild<ValuesOf<T>>>

export class ClientService extends BaseAPI  {
    constructor(client: Client, baseUrl: string, config: ClientConfig) {
        super(client, baseUrl, config);
    }

    async getTicketById(args: ArgsQuery['Ticket'], body: Body<Query['Ticket']>): Promise<Query['Ticket']> {
        try {
            const query = this.generateOperation(Operation.QUERY, 'Ticket', args, body)
            const { data } = await this.post(query)
            this.errorHandler(data)
            return data.data.Ticket
        } catch (e: any) {
            return Promise.reject(e)
        }
    }

    async getTickets(args: ArgsQuery['Tickets'], body: Body<Query['Tickets']>): Promise<Query['Tickets']> {
        try {
            const query = this.generateOperation(Operation.QUERY, 'Tickets', args, body)
            const { data } = await this.post(query)
            this.errorHandler(data)
            return data.data.Tickets
        } catch (e: any) {
            return Promise.reject(e)
        }
    }

    async suggestAnswer(args: ArgsQuery['SuggestAnswer'], body: Body<Query['SuggestAnswer']>): Promise<Query['SuggestAnswer']> {
        try {
            const query = this.generateOperation(Operation.QUERY, 'SuggestAnswer', args, body)
            const { data } = await this.post(query)
            this.errorHandler(data)
            // this.query()
            return data.data.SuggestAnswer
        } catch (e: any) {
            return Promise.reject(e)
        }
    }

    async getTicketDepartments(body: Body<Query['TicketDepartments']>): Promise<Query['TicketDepartments']> {
        try {
            const query = this.generateOperation(Operation.QUERY, 'TicketDepartments', {}, body)
            const { data } = await this.post(query)
            this.errorHandler(data)
            return data.data.TicketDepartments
        } catch (e: any) {
            return Promise.reject(e)
        }
    }

    private createFormDate(mutation: string, attachments: File): FormData {
        const bodyFormData = new FormData();
        bodyFormData.set('operations', JSON.stringify({
            query: `${mutation}`,
            variables: { attachments }
        })) 
        bodyFormData.set('operationName', '');
        bodyFormData.set('map', JSON.stringify({"file":["variables.file"]}));
        bodyFormData.append('file', attachments);
        return bodyFormData
    }

    async ticketAdd(args: ArgsMutation['TicketAdd'], body: Body<Mutation['TicketAdd']>): Promise<Mutation['TicketAdd']> {
        try {
            const mutation = this.generateOperation(Operation.MUTATION, 'TicketAdd', args, body)
            const { attachments } = args
            const queryData = attachments ? this.createFormDate(mutation, attachments) : mutation;
            // console.log('@@@@@')
            // console.log(mutation)
            // return 
            const { data } = await this.post(queryData)
            this.errorHandler(data)
            return data.data.TicketAdd
        } catch (e: any) {
            return Promise.reject(e)
        }
    }

    async ticketAddReply(args: ArgsMutation['TicketAddReply'], body: Body<Mutation['TicketAddReply']>): Promise<Mutation['TicketAddReply']> {
        try {
            const mutation = this.generateOperation(Operation.MUTATION, 'TicketAddReply', args, body)
            const { attachments } = args
            const queryData = attachments ? this.createFormDate(mutation, attachments) : mutation;
            const { data } = await this.post(queryData)
            this.errorHandler(data)
            return data.data.TicketAddReply
        } catch (e: any) {
            return Promise.reject(e)
        }
    }

    async ticketFeedback(args: ArgsMutation['TicketFeedback']): Promise<Mutation['TicketFeedback']> {
        try {
            const mutation = this.generateOperation(Operation.MUTATION, 'TicketFeedback', args, [])
            const { data } = await this.post(mutation)
            this.errorHandler(data)
            return data.data.TicketFeedback
        } catch (e: any) {
            return Promise.reject(e)
        }
    }

    async ticketReplyFeedback(args: ArgsMutation['TicketReplyFeedback']): Promise<Mutation['TicketReplyFeedback']> {
        try {
            const mutation = this.generateOperation(Operation.MUTATION, 'TicketReplyFeedback', args, [])
            const { data } = await this.post(mutation)
            this.errorHandler(data)
            return data.data.TicketReplyFeedback
        } catch (e: any) {
            return Promise.reject(e)
        }
    }

    async ticketClose(args: ArgsMutation['TicketClose']): Promise<Mutation['TicketClose']> {
        try {
            const mutation = this.generateOperation(Operation.MUTATION, 'TicketClose', args, [])
            const { data } = await this.post(mutation)
            this.errorHandler(data)
            return data.data.TicketClose
        } catch (e: any) {
            return Promise.reject(e)
        }
    }

    private generateArgs(args: any): string {
        const enums = ["LOW", "MEDIUM", "HIGH", "HAPPY", "NEUTRAL", "SAD", "NONE", "NEW", "IN_PROGRESS", "SUPPORT_RESPONSE", "CLOSED"] as const;
        type Enums = (typeof enums)[number];
        const isEnums = (x: any): x is Enums => enums.includes(x) ;
        let str =  ``
        for (let arg in args) {
            const value = args[arg]
            if (typeof value === "string") {
                str += `${arg}:${ isEnums(value) ? value : `"${value}"` },`;
            } else {
                str += `${arg}:${value},`
            }
        }
        return str
    }

    private generateOperation(
        type: Operation,
        name: OperationName,
        args: ValueOf<ArgsQuery> | ValueOf<ArgsMutation>,
        body: any[]): string  {
            let argument = args
            if (this.hasAttachment(args)) {
                const { attachments, ...restArgs } = args
                argument = restArgs
            }
            const generatedType = this.hasAttachment(args) ? `${type} ${name}($file: Upload!)` : `${type}`;
            const generatedArgs = `${name} ${Object.keys(args).length ? `(${this.generateArgs(argument)} ${this.hasAttachment(args) ? 'attachments: $file' : '' } )` : ''}`;
            const generatedBody = `${body.length ? `{ ${this.generateBody(body)} }` : this.generateBody(body)}`;
            return `${generatedType} { ${generatedArgs} ${generatedBody} }`
    }

    private hasAttachment(args: object): args is ArgsMutation['TicketAdd'] | ArgsMutation['TicketAddReply'] {
        return args.hasOwnProperty('attachments')
    }

    private generateBody(body: any[]): string {
        let str = ''
        for(let item of body) {
            if(typeof(item) === 'object') {
               for (let key in item){
                   str += `${key} { ${this.generateBody(item[key])} }`
               }
            } else {
                str += ` ${item} `
            }

        }
        return str
    }

    private errorHandler(data: {data?: any, errors?: any}) {
        if(data.errors) {
            const { errors } = data
            if(errors[0].validation?.length) {
                throw new ValidationError(errors[0].message, errors)
            } else {
                throw new ApplicationError(errors[0].message, errors)
            }
        }
    }
}
